module FFJ.FeatureModel where {
	import List;
	import Debug.Trace;

	type Feature = String;
	
	data Formula = Tru | Fls | Var Feature | And [Formula] | Or [Formula] | Not Formula | Impl Formula Formula deriving Show;
	
	type FM = [Formula];
	
	s1 = Impl (Var "Eval") (And [(Var "Add"), (Var "Sub")]);
	s2 = Impl (Var "Sub") (Var "Add");
	s3 = Var "Add";
	s4 = Impl (Var "Eval") (Not (Var "Eval2"));
	s5 = Impl (Var "Eval2") (And [(Var "Add"), (Var "Sub")]);
	
	fm = [s1,s2,s3,s4,s5];
	
	t1 = implies fm "Sub" ["Add"];
	t2 = implies fm "Add" ["Sub"];
	t3 = implies fm "Eval" ["Sub"];
	t4 = implies fm "Sub" ["Eval"];
	t5 = implies fm "Eval" ["Add"];
	t6 = implies fm "Add" ["Eval"];
	t7 = implies fm "Eval" ["Eval2"];
	t11 = present fm "Sub" ["Add"];
	t22 = present fm "Add" ["Sub"];
	t33 = present fm "Eval" ["Sub"];
	t44 = present fm "Sub" ["Eval"];
	t55 = present fm "Eval" ["Add"];
	t66 = present fm "Add" ["Eval"];
	t77 = present fm "Eval" ["Eval2"];


	present :: FM -> Feature -> [Feature] -> Bool;
	present fm phi psis = not (solve (Not (And (fm ++ [(Var phi)] ++ (getVars psis)))));
	
	implies :: FM -> Feature -> [Feature] -> Bool;
	implies fm phi psis = solve (Impl (And fm) (Impl (Var phi) (And (getVars psis))));

	getVars :: [Feature] -> [Formula];
	getVars (f:fs) = (Var f):(getVars fs);
	getVars [] = [];
	
	solve :: Formula -> Bool;
	
	solve f = case (findVar f) of {
		Just v -> solve (And [(assign v Tru f),(assign v Fls f)]);
		Nothing ->  boolean f;
	}; 
	
	findVar :: Formula -> Maybe Feature;
	
	findVar (Var n) = Just n;
	
	findVar (And (f:fs)) = case (findVar f) of {
		Just v -> Just v;
		Nothing -> findVar (And fs);
	};
	findVar (And []) = Nothing;

	findVar (Or (f:fs)) = case (findVar f) of {
		Just v -> Just v;
		Nothing -> findVar (Or fs);
	};
	findVar (Or []) = Nothing;
	
	findVar (Not f) = findVar f;
	
	findVar (Impl f g) = case (findVar f) of {
		Just v -> Just v;
		Nothing -> findVar g;
	};

	findVar Tru = Nothing;
	
	findVar Fls = Nothing;
	
	assign :: Feature -> Formula -> Formula -> Formula;
	assign n c (Var s) | (n == s) = c
                       | otherwise = (Var s);
    assign n c (And fs) =  (And (map (assign n c) fs));
    assign n c (Or fs) =  (Or (map (assign n c) fs));
    assign n c (Impl f1 f2) = (Impl (assign n c f1) (assign n c f2));
    assign n c (Not f) = (Not (assign n c f));
    assign n c (Tru) = Tru;
   	assign n c (Fls) = Fls;

   	
   	boolean :: Formula -> Bool;
   	boolean (And fs) = (foldl (&&) True (map boolean fs)); 
   	boolean (Or fs) = (foldl (||) False (map boolean fs));
   	boolean (Impl f1 f2) = ((not (boolean f1)) || (boolean f2));
   	boolean (Not f) = (not (boolean f));
   	boolean (Tru) = True;
   	boolean (Fls) = False;
    
}