
module FFJ.Print where {
	import FFJ.Program;
	import FFJ.Syntax;
	import Monad;

	printct :: CT -> IO();
	printct ct = do {
		mapM_ printclass ct;
	};

	printpt :: PT -> IO();
	printpt pt = do {
		mapM_ printprog pt;
	};

	printprog :: (PTEntry,Term) -> IO();
	printprog ((c,(x,feat)),t) = do {
		putStr (c ++ " " ++ feat ++ "::" ++ x ++ " = ");
		printterm t;
		putStr "\n";
	};

	printclass :: (ExtType, Class) -> IO();
	printclass ((e,feat), (o, c, d, f, m)) = do { 
		if o then putStr "refines " else putStr "";
		putStr "class ";
		putStr (feat ++ "::" ++ c);
		if (d /= "") && (not o) then putStr " extends " else putStr "";
		when (not o) (putStr d);
		putStr " {\n";
		mapM_ printfield f;
		mapM_ printmethod m;
		putStr "}\n";
	};
	
	printfields :: [Field] -> IO();
	printfields fs = mapM_ printfield fs;
	
	printfield :: Field -> IO();
	printfield (t,f) = do {
		putStr "  "; putStr t; putStr " "; putStr f; putStr ";\n";
	};
	
	printmethods :: [Method] -> IO();
	printmethods fs = mapM_ printmethod fs;
	
	printmethod :: Method -> IO();
	printmethod (o,r,m,p,t) = do {
		putStr "  "; if o then putStr "override " else putStr ""; putStr r; putStr " "; putStr m; putStr "("; printparams p; putStr ") { return "; printterm t; putStr "; }\n";
	};
	
	printparams :: [Param] -> IO();
	printparams (x:xs) = do {
		printparam x; if (length xs /= 0) then putStr ", " else putStr ""; printparams xs;
	};
	printparams [] = do {
		putStr "";
	};
	
	printparam :: Param -> IO();
	printparam (t,v) = do {
		putStr t; putStr " "; putStr v;
	};

	printterm :: Term -> IO();
	printterm (Var v) = do {
		putStr v;
	};
	printterm (Proj t f) = do {
		printterm t; putStr ("." ++ f);
	};
	printterm (Inv t m a) = do {
		printterm t; putStr ("." ++ m ++ "("); printargs a; putStr ")";
	};
	printterm (Inst c t) = do {
		putStr ("new " ++ c ++ "("); printargs t; putStr ")"; 
	};
	printterm (Cast c t) = do {
		putStr ("(" ++ c ++ ")"); printterm t;
	};
	
	printargs :: [Term] -> IO();
	printargs (x:xs) = do {
		printterm x; if (length xs /= 0) then putStr ", " else putStr ""; printargs xs;
	};
	printargs [] = do {
		putStr "";
	};
	
	printids :: [Identifier] -> IO();
	printids (x:xs) = do {
		putStr x; if (length xs /= 0) then putStr ", " else putStr ""; printids xs;
	};
	printids [] = do {
		putStr "";
	};
	
	
	printsig :: Signature -> IO();
	printsig (bs,b) = do {
		putStr "["; printids bs; putStr "] -> "; putStr b;
	};

	printbody :: MethodBody -> IO();
	printbody (vs,t) = do {
		putStr "(["; printids vs; putStr "], "; printterm t; putStr ")";
	};

	printinits :: [(Variable,Variable)] -> IO();
	printinits ((v1,v2):xs) = do {
		putStr ("this." ++ v1 ++ "=" ++ v2 ++ "; "); printinits xs;
	};
	printinits [] = do {
		putStr "";
	};

	termToString :: Term -> String;
	termToString (Var v) = v;
	termToString (Proj t f) = (termToString t) ++ "." ++ f;
	termToString (Inv t m a) = (termToString t) ++ "." ++ m ++ "(" ++ (argsToString a) ++ ")";
	termToString (Inst c t) = "new " ++ c ++ "(" ++ (argsToString t) ++ ")"; 
	termToString (Cast c t) = "(" ++ c ++ ")" ++ (termToString t);
	
	argsToString :: [Term] -> String;
	argsToString (x:xs) = (if (length xs /= 0) then (termToString x) ++ ", " else (termToString x)) ++ (argsToString xs);
	argsToString [] = "";
	

}
