
module FFJ.Program where {
	import FFJ.Syntax;
	import Maybe;
	import Debug.Trace;
	
	type Feature = Identifier;
	type ExtType = (Type,Feature);
	type CT = [(ExtType, Class)];
	
	classdef :: CT -> ExtType -> Maybe Class;
	classdef ct c = lookup c ct;

	initct :: CT;
	initct = [(("Object","None"), (False, "Object", "", [], []))];
	
	getFirst :: CT -> Type -> ExtType;
	getFirst (((d,psi),_):xs) c | (c == d) = (d,psi)
	                              | otherwise = getFirst xs c;
	getFirst [] c = error ("For type " ++ c ++ " no feature can be found! (getFirst)");
	
	type ExtId = (Identifier,Feature);
	type PTEntry = (Type,ExtId);
	type PT = [(PTEntry, Term)];
	
	initpt :: PT;
	initpt = [];
	
	subtypes :: CT -> [ExtType] -> [ExtType] -> Bool;
	subtypes ct (c:cs) (d:ds) = if (length cs == length ds) then  (subtype ct c d) && (subtypes ct cs ds) else False;
	subtypes ct [] [] = True;
	subtypes ct cs [] = False;
	subtypes ct [] ds = False;

	
	subtypesT :: CT -> [Type] -> [Type] -> Bool;
	subtypesT ct (c:cs) (d:ds) = if (length cs == length ds) then (subtype ct (getFirst ct c) (getFirst ct d)) && (subtypesT ct cs ds) else False;
	subtypesT ct [] [] = True;
	subtypesT ct cs [] = False;
	subtypesT ct [] ds = False;
	
	subtype :: CT -> ExtType -> ExtType -> Bool;
	subtype ct (c,phi) (d,psi) | (c == d) = True 
	               | (c == "Object") = False 
	               | otherwise = if (isJust cd) then (if (d == sc) then True else (subtype ct (getFirst ct sc) (d,psi))) else error ("Class " ++ c ++ " not found in class table! (subtype)")
	               where { cd = (classdef ct (c,phi)); sc = superclass (fromJust cd) };
	
	superclass :: Class -> Type;
	superclass (_,_,d,_,_) = d;
	
	getPred :: CT -> ExtType -> Maybe ExtType;
	getPred ct c = getPredInner ct [] c;
	
	getPredInner :: CT -> [ExtType] -> ExtType -> Maybe ExtType;
	getPredInner (((d,psi),_):xs) prevtypes (c,phi) | ((d == c) && (psi == phi)) = if ((length prevtypes) == 0) then Nothing else Just (head prevtypes)
	                                         | ((d == c ) && (psi /= phi)) = getPredInner xs ((d,psi):prevtypes) (c,phi)
	                                         | otherwise = getPredInner xs prevtypes (c,phi);
	getPredInner [] prevtypes (c,phi) = error ("Class " ++ phi ++ "::" ++ c ++ " not found in class table! (getPredInner)");
    
    getLast :: CT -> Type -> ExtType;
    getLast ct c = getLastInner ct [] c;
    
    getLastInner :: CT -> [ExtType] -> Type -> ExtType;
    getLastInner (((d,psi),_):xs) prevtypes c | (d == c) = getLastInner xs ((d,psi) : prevtypes) c
                                              | otherwise = getLastInner xs prevtypes c;
    getLastInner [] prevtypes c | ((length prevtypes) == 0) = error ("Class " ++ c ++ " not found in class table! (getLastInner)")
                                | otherwise = (head prevtypes);
	
	
}