module FFJPL.FeatureModel where {
	import FFJPL.Definitions;

	data Formula = Tru | Fls | FVar Feature | And [Formula] | Or [Formula] | Not Formula | Impl Formula Formula deriving Show;

	type FeatureModel = ([Feature],[Formula]);

	sometimes :: FeatureModel -> [Feature] -> Feature -> Bool; -- feature model interface
	sometimes (fs,fa) omegas psi = not (solve (Not (And (fa ++ (getFVars omegas) ++ [(FVar psi)]))));

	always :: FeatureModel -> [Feature] -> Feature -> Bool; -- feature model interface
	always (fs,fa) omegas psi = solve (Impl (And fa) (Impl (And (getFVars omegas)) (FVar psi)));

	alwaysOR :: FeatureModel -> [Feature] -> [Feature] -> Bool;
	alwaysOR (fs,fa) omegas psis = solve (Impl (And fa) (Impl (And (getFVars omegas)) (Or (getFVars psis))));

	never :: FeatureModel-> [Feature] -> Feature -> Bool; -- feature model interface
	never fm omegas psi = not (sometimes fm omegas psi);

	alwaysGroup :: FeatureModel -> [Feature] -> Feature -> [Feature];
	alwaysGroup fm omegas psi = if (always fm omegas psi) 
	                               then [psi]
	                               else if (alwaysOR fm omegas (psi:(neverGroup fm omegas psi)))
	                                    then (psi:(neverGroup fm omegas psi))
	                                    else [];
	neverGroup :: FeatureModel -> [Feature] -> Feature -> [Feature];
	neverGroup (fs,fa) omegas psi = neverGroupIter (fs,fa) fs omegas psi;

	neverGroupIter :: FeatureModel -> [Feature] -> [Feature] -> Feature -> [Feature];
	neverGroupIter fm (f:rest) omegas psi = if (never fm (f:omegas) psi) 
	                                    then (neverGroupIter fm rest omegas psi) ++ [f]
	                                    else (neverGroupIter fm rest omegas psi);
	neverGroupIter fm [] omegas psi = [];
	
	getFVars :: [Feature] -> [Formula];
	getFVars (f:fs) = (FVar f):(getFVars fs);
	getFVars [] = [];

	solve :: Formula -> Bool;
	
	solve f = case (findFVar f) of {
		Just v -> solve (And [(assign v Tru f),(assign v Fls f)]);
		Nothing ->  boolean f;
	}; 
	
	findFVar :: Formula -> Maybe Feature;
	
	findFVar (FVar n) = Just n;
	
	findFVar (And (f:fs)) = case (findFVar f) of {
		Just v -> Just v;
		Nothing -> findFVar (And fs);
	};
	findFVar (And []) = Nothing;

	findFVar (Or (f:fs)) = case (findFVar f) of {
		Just v -> Just v;
		Nothing -> findFVar (Or fs);
	};
	findFVar (Or []) = Nothing;
	
	findFVar (Not f) = findFVar f;
	
	findFVar (Impl f g) = case (findFVar f) of {
		Just v -> Just v;
		Nothing -> findFVar g;
	};

	findFVar Tru = Nothing;
	
	findFVar Fls = Nothing;
	
	assign :: Feature -> Formula -> Formula -> Formula;
	assign n c (FVar s) | (n == s) = c
                       | otherwise = (FVar s);
    assign n c (And fs) =  (And (map (assign n c) fs));
    assign n c (Or fs) =  (Or (map (assign n c) fs));
    assign n c (Impl f1 f2) = (Impl (assign n c f1) (assign n c f2));
    assign n c (Not f) = (Not (assign n c f));
    assign n c (Tru) = Tru;
   	assign n c (Fls) = Fls;

   	
   	boolean :: Formula -> Bool;
   	boolean (And fs) = (foldl (&&) True (map boolean fs)); 
   	boolean (Or fs) = (foldl (||) False (map boolean fs));
   	boolean (Impl f1 f2) = ((not (boolean f1)) || (boolean f2));
   	boolean (Not f) = (not (boolean f));
   	boolean (Tru) = True;
   	boolean (Fls) = False;
    
}