module FFJPL.Functions where {
	import FFJ.Syntax;
	import FFJPL.Definitions;
	import FFJPL.FeatureModel;
	import {-# SOURCE #-} FFJPL.Subtype;
	import Maybe;
	
	initCT :: CT;
	initCT = [(("Object","Base"), (False, "Object", "", [], []))];
	
	getClass :: CT -> ExtType -> Class;
	getClass ct (c,phi) = case (lookup (c,phi) ct) of {
		Just decl -> decl;
		Nothing -> error ("Type error: Qualified type " ++ phi ++ "." ++ c ++ " not found in class table! (getClass)")
	};
	
	getSuperClass :: CT -> ExtType -> Type;
	getSuperClass ct ce = getSuperClassType (getClass ct ce);
	
	getSuperClassType :: Class -> Type;
	getSuperClassType (_,_,d,_,_) = d;

	mdecl :: Identifier -> [Method] -> Maybe Method;
	mdecl n ((False,r,m,ps,b):xs) | (m == n) = Just (False,r,m,ps,b) | otherwise = mdecl n xs;
	mdecl n ((True,r,m,ps,b):xs) = mdecl n xs;
	mdecl n [] = Nothing;
	
	fdecl :: Identifier -> [Field] -> Maybe Field;
	fdecl g ((c,f):fs) | (f == g) = Just (c,f) | otherwise = fdecl g fs;
	fdecl g [] = Nothing;

	getIntroductions :: FeatureModel -> CT -> Type -> [Feature]; -- introduction table (IT)
	getIntroductions fm (((c,phi),(o,_,_,_,_)):next) e
		| ((c == e) && (not o)) = phi : getIntroductions fm next e
		| otherwise = getIntroductions fm next e;
	getIntroductions _ [] _ = [];
	
	getIntroductionsScoped :: FeatureModel -> CT -> [Feature] -> Type -> [Feature]; 
	getIntroductionsScoped fm (((c,phi),(o,_,_,_,_)):next) psis e
		| ((c == e) && (not o) && (any (==phi) psis)) = phi : getIntroductionsScoped fm next psis e
		| otherwise = getIntroductionsScoped fm next psis e;
	getIntroductionsScoped _ [] _ _ = [];

	getFieldIntroductions :: FeatureModel -> CT -> [Feature] -> (Type,Identifier) -> [Feature]; 
	getFieldIntroductions fm ct omegas fdef = getFieldIntroductionsIter fm ct ct omegas fdef;

	getFieldIntroductionsIter :: FeatureModel -> CT -> CT -> [Feature] -> (Type,Identifier) -> [Feature]; 
	getFieldIntroductionsIter fm ct (((c,phi),(_,_,_,fs,_)):next) omegas (e,f)
		| ((isSubType fm ct e c omegas) && (isJust (fdecl f fs))) = phi : getFieldIntroductionsIter fm ct next omegas (e,f)
		| otherwise = getFieldIntroductionsIter fm ct next omegas (e,f);
	getFieldIntroductionsIter _ _ [] _ _ = [];

	getMethodIntroductions :: FeatureModel -> CT -> [Feature] -> (Type,Identifier) -> [Feature]; 
	getMethodIntroductions fm ct omegas mdef = getMethodIntroductionsIter fm ct ct omegas mdef;

	getMethodIntroductionsIter :: FeatureModel -> CT -> CT -> [Feature] -> (Type,Identifier) -> [Feature]; 
	getMethodIntroductionsIter fm ct (((c,phi),(_,_,_,_,ms)):next) omegas (e,m)
		| ((isSubType fm ct e c omegas) && (isJust (mdecl m ms))) = phi : getMethodIntroductionsIter fm ct next omegas (e,m)
		| otherwise = getMethodIntroductionsIter fm ct next omegas (e,m);
	getMethodIntroductionsIter _ _ [] _ _ = [];


	getMethodIntroductionsScoped :: FeatureModel -> CT -> [Feature] -> [Feature] -> (Type,Identifier) -> [Feature]; 
	getMethodIntroductionsScoped fm ct omegas psis mdef = getMethodIntroductionsScopedIter fm ct ct omegas psis mdef;

	getMethodIntroductionsScopedIter :: FeatureModel -> CT -> CT -> [Feature] -> [Feature] -> (Type,Identifier) -> [Feature]; 
	getMethodIntroductionsScopedIter fm ct (((c,phi),(_,_,_,_,ms)):next) omegas psis (e,m)
		| ((isSubType fm ct e c omegas) && (isJust (mdecl m ms)) && (any (==phi) psis)) = phi : getMethodIntroductionsScopedIter fm ct next omegas psis (e,m)
		| otherwise = getMethodIntroductionsScopedIter fm ct next omegas psis (e,m);
	getMethodIntroductionsScopedIter _ _ [] _ _ _ = [];

	getRefinements :: FeatureModel -> CT -> Type -> [Feature]; -- refinement table (RT)
	getRefinements fm (((c,phi),classdec):next) e
		| (c == e) = phi : getRefinements fm next e
		| otherwise = getRefinements fm next e;
	getRefinements _ [] _ = [];
	
	isIntroduction :: CT -> ExtType -> Bool;
	isIntroduction ct ce = case (lookup ce ct) of {
	                         Just (False,_,_,_,_) -> True;
	                         Just (True,_,_,_,_) -> False;
	                         Nothing -> False;
	                       };
}