module FFJPL.Predicates where {
	import FFJ.Syntax;
	import FFJPL.Definitions;
	import FFJPL.Functions;
	import FFJPL.FeatureModel;
	import FFJPL.FieldLookup;
	import FFJPL.MethodLookup;
	import Maybe;
	
	validRef :: FeatureModel -> CT -> [Feature] -> Type -> Bool; 
	validRef fm ct omegas c = (c == "Object") || (alwaysOR fm omegas (getIntroductions fm ct c));
	
	validRefScoped :: FeatureModel -> CT -> [Feature] -> [Feature] -> Type -> Bool; 
	validRefScoped fm ct omegas psis c = alwaysOR fm omegas (getIntroductionsScoped fm ct psis c);

	validRefField :: FeatureModel -> CT -> [Feature] -> (Type,Identifier) -> Bool;
	validRefField fm ct omegas (c,f) = alwaysOR fm omegas (getFieldIntroductions fm ct omegas (c,f));

	validRefMethod :: FeatureModel -> CT -> [Feature] -> (Type,Identifier) -> Bool;
	validRefMethod fm ct omegas (c,m) = alwaysOR fm omegas (getMethodIntroductions fm ct omegas (c,m));

	validRefMethodScoped :: FeatureModel -> CT -> [Feature] -> [Feature] -> (Type,Identifier) -> Bool;
	validRefMethodScoped fm ct omegas psis (c,m) = alwaysOR fm omegas (getMethodIntroductionsScoped fm ct omegas psis (c,m));
	
	introduceClass :: FeatureModel -> CT -> [Feature] -> ExtType -> Bool;
	introduceClass (fs,fa) ct omegas ce = introduceClassIter (fs,fa) fs ct omegas ce;
	
	introduceClassIter :: FeatureModel -> [Feature] -> CT -> [Feature] -> ExtType -> Bool;	
	introduceClassIter fm (psi:fs) ct omegas (c,phi) | (psi /= phi) && (sometimes fm omegas psi) && (isIntroduction ct (c,psi)) = False
	                                               | otherwise = introduceClassIter fm fs ct omegas (c,phi);
	introduceClassIter fm [] ct omegas (c,phi) = True;
	
	introduceField :: FeatureModel -> CT -> [Feature] -> Identifier -> ExtType -> Bool;
	introduceField fm ct omegas f ce = and (map (\fs -> not (any ((==f) . snd) fs)) (fields fm ct omegas ce));

	introduceMethod :: FeatureModel -> CT -> [Feature] -> Identifier -> ExtType -> Bool;
	introduceMethod fm ct omegas m ce = (length (mtype fm ct omegas m ce) == 0);
	
	refine :: FeatureModel -> CT -> [Feature] -> ExtType -> Bool;
	refine fm ct omegas (c,phi) = validRefScoped fm ct omegas (fst (break (==phi) (getRefinements fm ct c))) c;
	
	override :: FeatureModel -> CT -> [Feature] -> Identifier -> ExtType -> Signature -> Bool;
	override fm ct omegas m (c,phi) (args,res) = (validRefMethodScoped fm ct omegas (phi:(fst (break (==phi) (getRefinements fm ct c)))) (c,m)) &&
	                                             (all (==(args,res)) (mtype fm ct omegas m (c,phi)));

	                                             
 }