module FFJPL.SanityConditions where {
	import FFJ.Syntax;
	import FFJPL.Definitions;
	import FFJPL.Functions;
	import FFJPL.FeatureModel;
	import List;

	sanityConditionsOK :: FeatureModel -> CT -> Bool;
	sanityConditionsOK fm ct = (uniqueBase fm) && (uniqueListElements ct ct) && (cycleFree fm ct ct);

	uniqueBase :: FeatureModel -> Bool;
	uniqueBase (features,_) = if (length (findIndices (== ("Base")) features) == 1) then True else error "Type error: Feature Base is not unique in the feature model! ";

--  It is not necessary to check whether a user-defined class exists that has the name Object => the parser rejects such declarations
--	objectOK :: CT -> Bool;
--	objectOK ct = case (lookup ("Object","Base") ct) of {
--		Just (o,c,d,f,m) -> if (((o == False) && (c == "Object") && (length f == 0) && (length m == 0) && (d == ""))) then True else error "Type error: Class Object is missing or not well-formed in the class table! ";
--		Nothing -> error "Type error: Class Object is missing or not well-formed in the class table! "; 
--	};
--
--	objectUnique :: CT -> Bool;
--	objectUnique ct = if (length (findIndices (== "Object") (fst (unzip (fst (unzip ct))))) == 1) then True else error "Type error: Class Object is not unique in the class table! ";

	cycleFree :: FeatureModel -> CT -> CT -> Bool;
	cycleFree fm ct ((c,cdecl):xs) = case cdecl of {
		(False,_,d,_,_) -> if (not (or (map (\f -> (sometimes fm [(snd c)] f) && (inducesCycle fm ct [(snd c)] c (d,f))) (getIntroductions fm ct d)))) then (cycleFree fm ct xs) else error ("Type error: Qualified type " ++ (snd c) ++ "." ++ (fst c) ++ " induces a cycle in the class table! ");
		(True,_,_,_,_) -> cycleFree fm ct xs;
	};
	cycleFree fm ct [] = True;
	
	inducesCycle :: FeatureModel -> CT -> [Feature] -> ExtType -> ExtType -> Bool;
	inducesCycle fm ct omegas (c,phi) (d,psi) | (c == "Object" && phi =="Base") || (d == "Object" && psi == "Base") = False
                                       | (c == d) = True 
	                                   | otherwise = (or (map (\f -> (sometimes fm (omegas++[psi]) f) && (inducesCycle fm ct (omegas++[psi]) (c,phi) (e,f))) (getIntroductions fm ct e)))
	                                   where { e = getSuperClass ct (d,psi); };


--	uniqueClassesAndRefinements :: FeatureModel -> CT -> Bool;
--	uniqueClassesAndRefinements (feature:features,formulas) ct =  ((snd ( unzip (fst (unzip ct))))) && (uniqueClassesAndRefinements (features,formulas));
	                                   
	uniqueListElements :: CT -> CT -> Bool;
	uniqueListElements ct (((c,phi),(_,_,_,fs,ms)):cdecls) = if ((uniqueClass ct (c,phi)) && (uniqueFields (c,phi) fs) && (uniqueMethods (c,phi) ms)) then (uniqueListElements ct cdecls) else False; 
	uniqueListElements _ [] = True;  
	
	uniqueClass :: CT -> ExtType -> Bool;
	uniqueClass ct (c,phi) = if (length (findIndices (== (c,phi)) (fst (unzip ct))) == 1) then True else error ("Type error: Feature " ++ phi ++ " contains multiple classes or refinements of type " ++ c ++ "!");
	
	uniqueFields :: ExtType -> [Field] -> Bool;
	uniqueFields (c,phi) fs = if (length fs == length (nubBy (\(c,f) (g,h) -> f == h) fs)) then True else error ("Type error: Qualified type " ++ phi ++ "." ++ c ++ " contains multiple fields with identical names!");
	
	uniqueMethods :: ExtType -> [Method] -> Bool;
	uniqueMethods (c,phi) ms = if (length ms == length (nubBy (\(o1,r1,m1,ps1,t1) (o2,r2,m2,ps2,t2) -> m1 == m2) ms)) then (all (\(_,_,m,ps,_) -> uniqueParams (c,phi) m ps) ms) else error ("Type error: Qualified type " ++ phi ++ "." ++ c ++ " contains multiple methods with identical names!");
	                                   
	uniqueParams :: ExtType -> Identifier -> [Param] -> Bool;
	uniqueParams (c,phi) m ps = if (length ps == length (nubBy (\(c1,v1) (c2,v2) -> v1 == v2) ps)) then True else error ("Type error: Method " ++ m ++ " of qualified type " ++ phi ++ "." ++ c ++ " contains multiple parameters with identical names!");
	                    
}   
	                 