
module FJ.Adapter where {
	import FJ.Syntax;
	import FJ.Program;
	import FJ.Parser;
	import Debug.Trace;
	
	getprog :: (CT,PT) -> [PStat] -> (CT,PT);
	getprog (ct,pt) (x:xs) = getprog (readstat (ct,pt) x) xs;
	getprog (ct,pt) [] = (ct,pt);
	
	
	readstat :: (CT,PT) -> PStat -> (CT,PT);
	readstat (ct,pt) (PClass c d pfc pmeth) = (ct++[(c,(c,(typename d),(getcons pfc),(getfields pfc),(getmethods pmeth)))], pt);
	readstat (ct,pt) (PDef c x t) = (ct,pt++[(((typename c),x), (term t))]);
	
	
	typename :: PCID -> Type;
	typename PObject = "Object";
	typename (PCID d) = d;
		
	getfields :: PFC -> [Field];
	getfields (PField (PDecl pcid f) pfield) = ((typename pcid), f) : (getfields pfield);
	getfields (PConstr _ _ _ _) = [];
	
	getcons :: PFC -> Cons;
	getcons (PConstr c ps sargs inits) = (c, (getparams ps), sargs, inits);
	getcons (PField _ pfield) = (getcons pfield);
	
	getmethods :: [PMethod] -> [Method];
	getmethods ((PMethod (PDecl r m) ps t):xs) = ((typename r),m,(getparams ps),(term t)) : (getmethods xs);
	getmethods [] = []; 
	
	getparams :: [PDecl] -> [Param];
	getparams ((PDecl t x):xs) = ((typename t), x) : (getparams xs);
	getparams [] = [];
	
	term :: PTerm -> Term;
	term (PVar PThis) = (Var "this"); 
	term (PVar (PVID s)) =  (Var s);
	term (PElim t ((PGet f):xs)) = connect (Proj (term t) f) xs;
	term (PElim t ((PCall m ts):xs)) = connect (Inv (term t) m (map term ts)) xs;
	term (PElim t []) = term t;
	term (PCast pcid t) = (Cast (typename pcid) (term t));
	term (PNew pcid ts) = (Inst (typename pcid) (map term ts));

	connect :: Term -> [PElim] -> Term;
	connect t ((PGet f):xs) = connect (Proj t f) xs;
	connect t ((PCall m ts):xs) = connect (Inv t m (map term ts)) xs;
	connect t [] = t;
}	
