
module FJ.Print where {
	import FJ.Program;
	import FJ.Syntax;

	printct :: CT -> IO();
	printct ct = do {
		mapM_ printclass ct;
	};

	printpt :: PT -> IO();
	printpt pt = do {
		mapM_ printprog pt;
	};

	printprog :: (PTEntry, Term) -> IO();
	printprog ((c,x),t) = do {
		putStr (c ++ " " ++ x ++ " = ");
		printterm t;
		putStr "\n";
	};

	printclass :: (CTEntry, Class) -> IO();
	printclass (_, (c, d, k, f, m)) = do { 
		putStr "class ";
		putStr c;
		if (d /= "") then putStr " extends " else putStr "";
		putStr d;
		putStr " {\n";
		mapM_ printfield f;
		if (c /= "Object") then printcons k else putStr "";
		mapM_ printmethod m;
		putStr "}\n";
	};
	
	printfield :: Field -> IO();
	printfield (t,f) = do {
		putStr "  "; putStr t; putStr " "; putStr f; putStr ";\n";
	};

	printcons :: Cons -> IO();
	printcons (c,ps,sargs,inits) = do {
		putStr ("  " ++ c ++ "("); 
		printparams ps; 
		putStr (") { super("); 
		printids sargs; 
		putStr "); ";
		printinits inits;
		putStr "}\n";
	};

	printmethod :: Method -> IO();
	printmethod (r,m,p,t) = do {
		putStr "  "; putStr r; putStr " "; putStr m; putStr "("; printparams p; putStr ") { return "; printterm t; putStr "; }\n";
	};
	
	printparams :: [Param] -> IO();
	printparams (x:xs) = do {
		printparam x; if (length xs /= 0) then putStr ", " else putStr ""; printparams xs;
	};
	printparams [] = do {
		putStr "";
	};
	
	printparam :: Param -> IO();
	printparam (t,v) = do {
		putStr t; putStr " "; putStr v;
	};

	printterm :: Term -> IO();
	printterm (Var v) = do {
		putStr v;
	};
	printterm (Proj t f) = do {
		printterm t; putStr ("." ++ f);
	};
	printterm (Inv t m a) = do {
		printterm t; putStr ("." ++ m ++ "("); printargs a; putStr ")";
	};
	printterm (Inst c t) = do {
		putStr ("new " ++ c ++ "("); printargs t; putStr ")"; 
	};
	printterm (Cast c t) = do {
		putStr ("(" ++ c ++ ")"); printterm t;
	};
	
	printargs :: [Term] -> IO();
	printargs (x:xs) = do {
		printterm x; if (length xs /= 0) then putStr ", " else putStr ""; printargs xs;
	};
	printargs [] = do {
		putStr "";
	};
	
	printids :: [Identifier] -> IO();
	printids (x:xs) = do {
		putStr x; if (length xs /= 0) then putStr ", " else putStr ""; printids xs;
	};
	printids [] = do {
		putStr "";
	};
	
	
	printsig :: Signature -> IO();
	printsig (bs,b) = do {
		putStr "["; printids bs; putStr "] -> "; putStr b;
	};

	printbody :: MethodBody -> IO();
	printbody (vs,t) = do {
		putStr "(["; printids vs; putStr "], "; printterm t; putStr ")";
	};

	printinits :: [(Variable,Variable)] -> IO();
	printinits ((v1,v2):xs) = do {
		putStr ("this." ++ v1 ++ "=" ++ v2 ++ "; "); printinits xs;
	};
	printinits [] = do {
		putStr "";
	};

	termToString :: Term -> String;
	termToString (Var v) = v;
	termToString (Proj t f) = (termToString t) ++ "." ++ f;
	termToString (Inv t m a) = (termToString t) ++ "." ++ m ++ "(" ++ (argsToString a) ++ ")";
	termToString (Inst c t) = "new " ++ c ++ "(" ++ (argsToString t) ++ ")"; 
	termToString (Cast c t) = "(" ++ c ++ ")" ++ (termToString t);
	
	argsToString :: [Term] -> String;
	argsToString (x:xs) = (if (length xs /= 0) then (termToString x) ++ ", " else (termToString x)) ++ (argsToString xs);
	argsToString [] = "";
	

}
