
module FJ.Semantics where {
	import FJ.Syntax;
	import FJ.Program;
	import FJ.AuxFunc;
	import FJ.Print;
	import List;
	import Maybe;

	evalpt :: CT -> PT -> PT;
	evalpt ct ((e,t):xs) = (e,(eval ct t)):(evalpt ct xs);
	evalpt ct [] = [];

	substpt :: PT -> PT;
	substpt (((c,x),t):xs) = ((c,x),t):(substpt (substptinner x (Cast c t) xs));
	substpt [] = [];
		
	substptinner :: Variable -> Term -> PT -> PT;
	substptinner v st (((c,x),t):ps) = ((c,x),(subst v st t)):(substptinner v st ps);
	substptinner v st [] = [];
		
	eval :: CT -> Term -> Term;
	eval ct (Proj (Inst c ts) f) | (values ts) = let idx = (elemIndex f (snd (unzip (fields ct c)))) in if (isJust idx) then eval ct (ts !! (fromJust idx)) else error ("Runtime error: field " ++ f ++ " does not exist!");
	eval ct (Inv (Inst c ts) m us) | (values ts) = let body = (mbody ct m c) in if (isJust body) then eval ct (substall (fst (fromJust body)) us (subst "this" (Inst c ts) (snd (fromJust body)))) else error ("Runtime error: method body " ++ m ++ " does not exist!");
	eval ct (Cast d (Inst c ts)) | ((values ts) && (subtype ct c d)) = (Inst c ts);
	eval ct (Proj t f) | ((value t) == False) = eval ct (Proj (eval ct t) f);
	eval ct (Inv t m ts) | ((value t) == False) = eval ct (Inv (eval ct t) m ts);
	eval ct (Inv v m ts) | (value v) = eval ct (Inv v m (map (eval ct) ts));
	eval ct (Inst c ts) = (Inst c (map (eval ct) ts));
	eval ct (Cast c t) | ((value t) == False) = eval ct (Cast c (eval ct t));
	eval ct t = case t of {
		(Cast c t) -> error ("Evaluation error: downcast (" ++ c ++ ")(" ++ (termToString t) ++ ") is stuck!");
		otherwise -> error ("Evaluation error: term " ++ (termToString t) ++ " is stuck!");
	};

	subst :: Variable -> Term -> Term -> Term;
	subst v t2 (Var w) | (v == w) = t2 | otherwise = (Var w);
	subst t1 t2 (Proj t f) = (Proj (subst t1 t2 t) f);
	subst t1 t2 (Inv t m a) = (Inv (subst t1 t2 t) m (map (subst t1 t2) a));
	subst t1 t2 (Inst c t) = (Inst c (map (subst t1 t2) t));
	subst t1 t2 (Cast c t) = (Cast c (subst t1 t2 t));

	substall :: [Variable] -> [Term] -> Term -> Term;
	substall (x:xs) (y:ys) t = substall xs ys (subst x y t);
	substall [] [] t = t;

	values :: [Term] -> Bool;
	values (x:xs) = (value x) && values xs;
	values [] = True;
	
	value :: Term -> Bool;
	value (Inst _ []) = True;
	value (Inst _ ts) = values ts; 
	value (Proj _ _) = False;
	value (Inv _ _ _) = False;
	value (Cast _ _) = False;
}